#include "TAreaList.h"
#include "Macros.h"
#include <stdio.h>
#include <algorithm>
#include <vector>
#include "TAreaListEntry.h"
#include "TMaxSeverity.h"
#include "DereferencedFunction.h"



typedef vector<TAreaListEntry *> AreasList;



// The following class defines a global interface for loading the file area
// definitions from the various BBS systems.
// The implement support for a new BBS system you have to derive from this
// class and implement (at least) the virtual methods.

class TAreaConfig
{
  public:
    TAreaConfig (const TAreaSort & sort)
    : fSort (sort) {}
    virtual ~TAreaConfig () {}

    virtual void locateToTop () = 0;
    virtual TAreaListEntry * locateNextRecord (string & name, bool & descOnly) = 0;

    virtual bool getArea (struct TArea & inf, const TAreaListEntry & entry) = 0;

    virtual bool check () = 0;

  protected:
    TAreaSort fSort;
};



// This is the implementation for Maximus CBCS v2 of TAreaConfig

class TMaxAreaConfig_v2 : public TAreaConfig
{
  public:
    TMaxAreaConfig_v2 (const string & path, const TAreaSort & sort);
    virtual ~TMaxAreaConfig_v2 ();

    virtual void locateToTop ();
    virtual TAreaListEntry * locateNextRecord (string & name, bool & descOnly);

    virtual bool getArea (struct TArea & inf, const TAreaListEntry & entry);

    virtual bool check () { return fAreaFile != 0; }

  private:
    enum AREA
    {
      LEN       = 4,
      NAME      = 8,
      PATH      = 398,
      FILESBBS  = 638,
      INFO      = 718,
      PRIV      = 798,
      AREA_SIZE = 800
    };

    FILE *fAreaFile;
    long fLocation;
};

TMaxAreaConfig_v2::TMaxAreaConfig_v2 (const string & path, const TAreaSort & sort)
: TAreaConfig (sort), fAreaFile (0), fLocation (0)
{
  char *filename;

  filename = new char[path.length () + 10];
  strcpy (filename, path.c_str ());
  if (filename[path.length () - 1] != PATH_DELIMITER_CHAR)
    strcat (filename, PATH_DELIMITER);
  strcat (filename, "AREA.DAT");

  fAreaFile = fopen (filename, "rb");

  char id[4];
  if ((fread (id, 4, 1, fAreaFile) == 0) ||
      (id[0] != 'A') || (id[1] != 'D') || (id[2] != 'A') || (id[3] != 'T'))
  {
    fclose (fAreaFile);
    fAreaFile = 0;
  }
}

TMaxAreaConfig_v2::~TMaxAreaConfig_v2 ()
{
  if (fAreaFile)
    fclose (fAreaFile);
}

void TMaxAreaConfig_v2::locateToTop ()
{
  fLocation = 0;
}

TAreaListEntry * TMaxAreaConfig_v2::locateNextRecord (string & name, bool & descOnly)
{
  unsigned char buffer[AREA_SIZE];    // read only required number of bytes
  TAreaListEntry *entry;

  entry = 0;
  fseek (fAreaFile, fLocation, SEEK_SET);
  while ((fread (buffer, AREA_SIZE, 1, fAreaFile) != 0) && (entry == 0))
  {
    if (buffer[PATH] != '\0')                    // contains valid path
    {
      char *str;
      char *ptr;

      str = (char *) &buffer[NAME];
      for (ptr = str; *ptr != '\0'; ptr++)
        *ptr = toupper (*ptr);
      name = str;

      entry = new TAreaListEntry (name, fLocation, fSort);
    }

    fLocation += buffer[LEN] + (buffer[LEN + 1] << 8);

    if (!entry)
      fseek (fAreaFile, fLocation, SEEK_SET);
  }

  descOnly = false;

  return entry;
}

bool TMaxAreaConfig_v2::getArea (struct TArea & inf, const TAreaListEntry & entry)
{
  char buffer[AREA_SIZE];   // read only required number of bytes

  fseek (fAreaFile, entry.offset (), SEEK_SET);
  if (fread (buffer, AREA_SIZE, 1, fAreaFile) != 0)
  {
    inf.fPath = &buffer[PATH];
    if (inf.fPath[inf.fPath.length () - 1] != PATH_DELIMITER_CHAR)
      inf.fPath += PATH_DELIMITER;
    inf.fInfo = &buffer[INFO];
    if (buffer[638] == '\0')
      inf.fList = inf.fPath + "files.bbs";
    else
      inf.fList = &buffer[FILESBBS];
    inf.fPriv = TMaxSeverity::severity2ascii (((unsigned char *) buffer)[PRIV] +
                                              (((unsigned char *) buffer)[PRIV + 1] << 8));

    return true;
  }

  return false;
}



// This is the implementation for Maximus CBCS v3 of TAreaConfig

class TMaxAreaConfig_v3 : public TAreaConfig
{
  public:
    TMaxAreaConfig_v3 (const string & path, const TAreaSort & sort);
    virtual ~TMaxAreaConfig_v3 ();

    virtual void locateToTop ();
    virtual TAreaListEntry * locateNextRecord (string & name, bool & descOnly);

    virtual bool getArea (struct TArea & inf, const TAreaListEntry & entry);

    virtual bool check () { return fAreaFile != 0; }

  private:
    enum FAREA
    {
      LEN             = 0,
      NUM_OVER        = 2,
      CB_HEAP         = 4,
      ACS_OFFSET      = 8,
      NAME_OFFSET     = 10,
      DOWNPATH_OFFSET = 12,
      FILESBBS_OFFSET = 16,
      DESCRIPT_OFFSET = 18,
      ATTRIBS_OFFSET  = 32,
      FAREA_SIZE      = 64
    };

    FILE *fAreaFile;
    long fLocation;
};

TMaxAreaConfig_v3::TMaxAreaConfig_v3 (const string & path, const TAreaSort & sort)
: TAreaConfig (sort), fAreaFile (0), fLocation (0)
{
  char *filename;

  filename = new char[path.length () + 11];
  strcpy (filename, path.c_str ());
  if (filename[path.length () - 1] != PATH_DELIMITER_CHAR)
    strcat (filename, PATH_DELIMITER);
  strcat (filename, "FAREA.DAT");

  fAreaFile = fopen (filename, "rb");

  unsigned char id[4];
  if ((fread (id, 4, 1, fAreaFile) == 0) ||
      (id[0] != 0x3A) || (id[1] != 0x95) || (id[2] != 0x01) || (id[3] != 0x1A))
  {
    fclose (fAreaFile);
    fAreaFile = 0;
  }
}

TMaxAreaConfig_v3::~TMaxAreaConfig_v3 ()
{
  if (fAreaFile)
    fclose (fAreaFile);
}

void TMaxAreaConfig_v3::locateToTop ()
{
  fLocation = 4;
}

TAreaListEntry * TMaxAreaConfig_v3::locateNextRecord (string & name, bool & descOnly)
{
  unsigned char buffer[FAREA_SIZE];
  TAreaListEntry *entry;
  long heapsize;

  descOnly = false;
  entry = 0;
  fseek (fAreaFile, fLocation, SEEK_SET);
  while ((fread (buffer, FAREA_SIZE, 1, fAreaFile) != 0) && (entry == 0))
  {
    heapsize = buffer[CB_HEAP] + (buffer[CB_HEAP + 1] << 8);

    bool downpathAvail = (buffer[DOWNPATH_OFFSET] != 0) || (buffer[DOWNPATH_OFFSET + 1] != 0);
    descOnly = !downpathAvail;
    bool descFlag = (buffer[ATTRIBS_OFFSET] == 0) && (buffer[ATTRIBS_OFFSET + 1] == 0x40);

    if (downpathAvail || descFlag)
    {                                                  // contains valid path
      char *heap;
      char *ptr;
      char *str;

      heap = new char[heapsize];
      fread (heap, 1, heapsize, fAreaFile);
      str = &heap[buffer[NAME_OFFSET] + (buffer[NAME_OFFSET + 1] << 8)];
      for (ptr = str; *ptr != '\0'; ptr++)
        *ptr = toupper (*ptr);
      name = str;

      entry = new TAreaListEntry (name, fLocation, fSort);
    }

    fLocation += heapsize + buffer[LEN] + (buffer[LEN + 1] << 8) +
                 (buffer[NUM_OVER] + (buffer[NUM_OVER + 1] << 8)) * 8;

    if (!entry)
      fseek (fAreaFile, fLocation, SEEK_SET);
  }

  return entry;
}

bool TMaxAreaConfig_v3::getArea (struct TArea & inf, const TAreaListEntry & entry)
{
  unsigned char buffer[FAREA_SIZE];   // read only required number of bytes

  fseek (fAreaFile, entry.offset (), SEEK_SET);
  if (fread (buffer, FAREA_SIZE, 1, fAreaFile) != 0)
  {
    long heapsize = buffer[CB_HEAP] + (buffer[CB_HEAP + 1] << 8);
    char *heap = new char[heapsize];
    long offset;

    if (fread (heap, heapsize, 1, fAreaFile) == 0)
      return false;

    inf.fPath = &heap[buffer[DOWNPATH_OFFSET] + (buffer[DOWNPATH_OFFSET + 1] << 8)];
    if (inf.fPath[inf.fPath.length () - 1] != PATH_DELIMITER_CHAR)
      inf.fPath += PATH_DELIMITER;
    inf.fInfo = &heap[buffer[DESCRIPT_OFFSET] + (buffer[DESCRIPT_OFFSET + 1] << 8)];
    offset = buffer[FILESBBS_OFFSET] + (buffer[FILESBBS_OFFSET + 1] << 8);
    if (heap[offset] == '\0')
      inf.fList = inf.fPath + "files.bbs";
    else
      inf.fList = &heap[offset];
    inf.fPriv = &heap[buffer[ACS_OFFSET] + (buffer[ACS_OFFSET + 1] << 8)];

    delete [] heap;

    return true;
  }

  return false;
}



// Now the implementation of the class for which this file stands for

TAreaList::TAreaList (const string & path, bool mustSort, const TAreaSelectionList & fcmp,
                      const TAdditionalAreaList & addAreas, const TAreaSort & defSort)
{
  AreasList collect;

  fVersion = 0;

  // First collect the areas defined in a BBS configuration file

  fAreaConfig = new TMaxAreaConfig_v3 (path, defSort);
  if (!fAreaConfig->check ())
  {
    delete fAreaConfig;

    fAreaConfig = new TMaxAreaConfig_v2 (path, defSort);
    if (!fAreaConfig->check ())
    {
      delete fAreaConfig;
      fAreaConfig = 0;
    }
    else
    {
      fVersion = 2;

      cout << "Reading " << path << "AREA.DAT (Maximus v2)" << endl;
    }
  }
  else
  {
    fVersion = 3;

    cout << "Reading " << path << "FAREA.DAT (Maximus v3)" << endl;
  }

  if (fAreaConfig)                 // any maximus area configuration found ?
  {
    TAreaListEntry *entry;
    string areaname;
    bool descOnly;

    fAreaConfig->locateToTop ();

    while ((entry = fAreaConfig->locateNextRecord (areaname, descOnly)) != 0)
    {
      TArea area;

      fAreaConfig->getArea (area, *entry);

      fDescMap.insert (descMap::value_type (area.fName, area.fInfo));

      if (!descOnly)
      {
        TAreaSort areasort = entry->sortInfo ();

        if (fcmp.empty () || fcmp.matchArea (areaname, areasort))
          collect.push_back (new TAreaListEntry (areaname, entry->offset (), areasort));
      }

      delete entry;
    }
  }

  // Then add the areas defined in the configuration file

  if (addAreas.size () != 0)
    cout << "Appending file areas from configuration file" << endl;

  TAdditionalAreaList::const_iterator addIter = addAreas.begin ();
  while (addIter != addAreas.end ())
  {
    fDescMap.insert (descMap::value_type ((*addIter).getAreaName (),
                                          (*addIter).getFileDesc ()));
    TAreaSort addsort = defSort;

    if ((*addIter).getFilePath ().length () != 0)
    {
      if (fcmp.empty () || fcmp.matchArea ((*addIter).getAreaName (), addsort))
      {
        TAreaListEntry *newEntry = new TAreaListEntry ((*addIter).getAreaName (), -1, addsort);

        // check whether to overrule the info from Maximus definition
        AreasList::iterator found = find (collect.begin (), collect.end (), newEntry);

        if (found != collect.end ())
        {
          cout << "Warning: Overruling area '" << (*addIter).getAreaName () << "' by definition in config file !" << endl;
          *(*found) = *newEntry;
          delete newEntry;
        }
        else
          collect.push_back (newEntry);
      }
    }

    ++addIter;
  }

  // At least put the result into the requested order (sorting)

  if (mustSort)
  {
    ::sort (collect.begin (), collect.end (), dereferenced_greater<TAreaListEntry>());

    AreasList::iterator sortIter = collect.begin ();
    while (sortIter != collect.end ())
    {
      push_back ((*sortIter));

      ++sortIter;
    }
  }
  else
    if (!fcmp.empty ())
    {
      TAreaSelectionList::const_iterator iterSel = fcmp.begin ();
      while (iterSel != fcmp.end ())
      {
        AreasList::iterator iterAreas = collect.begin ();
        while (iterAreas != collect.end ())
        {
          if ((*iterSel).match ((*iterAreas)->name ()) &&
              !(*iterSel).excluded ())
          {
            push_back ((*iterAreas));

            (*iterAreas)->setHandled ();
          }

          ++iterAreas;
        }

        ++iterSel;
      }

      AreasList::iterator iterKill = collect.begin ();
      while (iterKill != collect.end ())
      {
        if (!((*iterKill)->isHandled ()))
        {
          delete (*iterKill);

          collect.erase (iterKill);    // cleanup not copied entries - should never appear
        }

        ++iterKill;
      }
    }
}



TAreaList::~TAreaList ()
{
  while (begin () != end ())
  {
    delete (*(begin ()));

    erase (begin ());
  }

  delete fAreaConfig;
}



void TAreaList::locateFirstArea ()
{
  fIterator = begin ();
}



bool TAreaList::getArea (struct TArea & inf, const TAdditionalAreaList & addAreas)
{
  bool result;

  if (fIterator == end ())
    return false;

  inf.fName = (*fIterator)->name ();
  inf.fSortInfo = (*fIterator)->sortInfo ();

  if ((*fIterator)->offset () < 0)
  {                        // locate the area in the additional areas list
    TAdditionalArea compare ((*fIterator)->name ().c_str ());

    TAdditionalAreaList::const_iterator addIter = find (addAreas.begin (), addAreas.end (), compare);

    inf.fPath = (*addIter).getFilePath ();
    inf.fInfo = (*addIter).getFileDesc ();
    inf.fList = (*addIter).getFilesBbs ();
    inf.fPriv = TMaxSeverity::severity2ascii ((*addIter).getSeverity ());

    result = true;
  }
  else
    result = fAreaConfig->getArea (inf, *(*fIterator));

  ++fIterator;

  return result;
}



string TAreaList::divisionDescription (const string & area) const
{
  descMap::const_iterator findDesc = fDescMap.find (area);
  if (findDesc != fDescMap.end ())
    return (*findDesc).second;

  return "";
}

